package govaluate

import (
	"fmt"
	"errors"
	"bytes"
	"strconv"
	"unicode"
)

func parseTokens(expression string) ([]ExpressionToken, error) {

	var ret []ExpressionToken
	var token, lastToken ExpressionToken
	var state lexerState
	var stream *lexerStream
	var err error
	var found bool

	state = validLexerStates[0];
	stream = newLexerStream(expression);

	for stream.canRead() {

		token, err, found = readToken(stream);

		if(err != nil) {
			return ret, err	
		}

		if(!found) {
			break;
		}

		if(!state.canTransitionTo(token.Kind)) {

			firstStateName := fmt.Sprintf("%s [%v]", GetTokenKindString(state.kind), lastToken.Value);
			nextStateName := fmt.Sprintf("%s [%v]", GetTokenKindString(token.Kind), token.Value);

			return ret, errors.New("Cannot transition token types from " + firstStateName + " to " + nextStateName);
		}

		// append this valid token, find new lexer state.		
		ret = append(ret, token)
		
		for _, possibleState := range validLexerStates {
			
			if(possibleState.kind == token.Kind) {
				
				state = possibleState
				break;
			}
		}

		lastToken = token;
	}

	if(!state.isEOF) {
		return ret, errors.New("Unexpected end of expression");
	}

	return ret, nil
}

func readToken(stream *lexerStream) (ExpressionToken, error, bool) {

	var ret ExpressionToken
	var tokenValue interface{}
	var tokenString string
	var kind TokenKind
	var character rune
	var found bool

	// numeric is 0-9, or .
	// string starts with '
	// variable is alphanumeric, always starts with a letter
	// symbols are anything non-alphanumeric
	// all others read into a buffer until they reach the end of the stream
	for(stream.canRead()) {

		character = stream.readCharacter()

		if(unicode.IsSpace(character)) {
			continue
		}

		kind = UNKNOWN;

		// numeric constant
		if(isNumeric(character)) {

			tokenString = readTokenUntilFalse(stream, isNumeric);
			tokenValue, _ = strconv.ParseFloat(tokenString, 64)
			kind = NUMERIC;
			break;
		}
		
		// variable
		if(unicode.IsLetter(character)) {

			tokenValue = readTokenUntilFalse(stream, unicode.IsLetter);
			kind = VARIABLE;

			if(tokenValue == "true") {

				kind = BOOLEAN
				tokenValue = true
			} else {

				if(tokenValue == "false") {

					kind = BOOLEAN	
					tokenValue = false
				}
			}
			break;
		}

		if(!isNotQuote(character)) {
			tokenValue = readUntilFalse(stream, true, false, isNotQuote);
			kind = STRING;

			// advance the stream one position, since reading until false assumes the terminator is a real token
			stream.rewind(-1)
			break;
		}

		if(character == '(') {
			tokenValue = character
			kind = CLAUSE
			break;
		}

		if(character == ')') {
			tokenValue = character
			kind = CLAUSE_CLOSE
			break;
		}

		// must be a known symbol
		tokenString = readTokenUntilFalse(stream, isNotAlphanumeric);
		tokenValue = tokenString

		_, found = MODIFIER_SYMBOLS[tokenString];
		if(found) {

			kind = MODIFIER;
			break;
		}

		_, found = LOGICAL_SYMBOLS[tokenString];
		if(found) {

			kind = LOGICALOP;
			break;
		}

		_, found = COMPARATOR_SYMBOLS[tokenString];
		if(found) {

			kind = COMPARATOR;
			break;
		}

		errorMessage := fmt.Sprintf("Invalid token: %v", tokenValue);
		return ret, errors.New(errorMessage), false;
	}

	ret.Kind = kind;
	ret.Value = tokenValue;

	return ret, nil, (kind != UNKNOWN);
}

func readTokenUntilFalse(stream *lexerStream, condition func(rune)(bool)) string {

	var ret string;

	stream.rewind(1)
	ret = readUntilFalse(stream, false, true, condition);
	return ret;
}

func readUntilFalse(stream *lexerStream, includeWhitespace bool, breakWhitespace bool, condition func(rune)(bool)) string {

	var tokenBuffer bytes.Buffer
	var character rune

	for(stream.canRead()) {

		character = stream.readCharacter()

		if(unicode.IsSpace(character)) {

			if(breakWhitespace && tokenBuffer.Len() > 0) {
				break;
			}
			if(!includeWhitespace) {
				continue;
			}
		}

		if(condition(character)) {
			tokenBuffer.WriteString(string(character));
		} else {
			stream.rewind(1)
			break;
		}
	}

	return tokenBuffer.String();
}

func isNumeric(character rune) bool {
	return unicode.IsDigit(character) || character == '.';
}
func isNotQuote(character rune) bool {
	return character != '\'' && character != '"';
}
func isNotAlphanumeric(character rune) bool {
	return !(unicode.IsDigit(character) || 
		unicode.IsLetter(character) || 
		character == '(' || 
		character == ')' ||
		!isNotQuote(character));
}
